<?php

namespace IMATHUZH\Qfq\Core\Form\FormElement;

use IMATHUZH\Qfq\Core\Form\Form;
use IMATHUZH\Qfq\Core\Helper\Logger;
use IMATHUZH\Qfq\Core\Helper\Support;

class TemplateGroupFormElement extends ContainerFormElement
{
    public string $addButtonId;
    public string $qfqFieldsName;
    public string $templateName;
    public string $targetName;


    public function __construct($attributes, ?Form $form = null)
    {
        parent::__construct($attributes, $form);

        $this->handleDefaultValues();
        $this->handleElementIdsAndNames();

        // Nothing else to do if TG has no children
        if (count($this->formElements) == 0) return;

        $this->handleMaxCopies();
        $this->handleHtmlAttributes();

    }

    /**
     * Prepare ids and names for special html elements of the template group
     */
    private function handleElementIdsAndNames() : void {
        $this->addButtonId = 'add_button_' . $this->attributes[FE_ID];
        $this->qfqFieldsName = 'qfq_fields_' . $this->attributes[FE_ID]; // ='qfq-fields'
        $this->templateName = 'template_' . $this->attributes[FE_ID]; // ='template'
        $this->targetName = 'target_' . $this->attributes[FE_ID]; // ='template'
    }


    /**
     * Set default values for template group specific attributes (unless the are specified by the user)
     */
    private function handleDefaultValues() : void {
        Support::setIfNotSet($this->attributes, FE_TEMPLATE_GROUP_ADD_CLASS, 'btn btn-default');
        Support::setIfNotSet($this->attributes, FE_TEMPLATE_GROUP_ADD_TEXT, 'Add');
        Support::setIfNotSet($this->attributes, FE_TEMPLATE_GROUP_REMOVE_CLASS, 'btn btn-default');
        Support::setIfNotSet($this->attributes, FE_TEMPLATE_GROUP_REMOVE_TEXT, 'Remove');
        Support::setIfNotSet($this->attributes, FE_TEMPLATE_GROUP_CLASS);
        Support::setIfNotSet($this->attributes, FE_MAX_LENGTH, '5', '');
    }

    /**
     * Child FE's of type checkbox need the NAME_TG_COPIES attribute for some reason
     */
    private function handleMaxCopies(){
        foreach ($this->formElements as $fe) {
            if ($fe->attributes[FE_TYPE] == FE_TYPE_CHECKBOX) {
                $fe->attributes[NAME_TG_COPIES] = $this->attributes[FE_MAX_LENGTH];
            }
        }
    }

    /**
     * Prepares $this->htmlAttributes with all attributes for the TG's main wrapping div
     */
    private function handleHtmlAttributes() : void {
        $this->htmlAttributes['class'] = $this->qfqFieldsName;
        $this->htmlAttributes[HTML_ATTR_ID] = $this->targetName;
        $this->htmlAttributes[ATTRIBUTE_DATA_REFERENCE] = $this->attributes[FE_DATA_REFERENCE];
        $this->htmlAttributes['data-qfq-line-template'] = '#' . $this->templateName;
        $this->htmlAttributes['data-qfq-line-add-button'] = '#' . $this->addButtonId;
    }


    /**
     * Build real FormElements based on the current templateGroup FormElements in $this->feSpecNative.
     * Take a templateGroup 'copy', if at least one of the elements is filled with a non default value.
     * Stop filling elements if there are no further elements filled.
     *
     * @param int $max
     * @param string $htmlDelete
     * @param array $json
     *
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \DownloadException
     * @throws \PhpOffice\PhpSpreadsheet\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Reader\Exception
     * @throws \PhpOffice\PhpSpreadsheet\Writer\Exception
     * @throws \Twig\Error\LoaderError
     * @throws \Twig\Error\RuntimeError
     * @throws \Twig\Error\SyntaxError
     * @throws \UserFormException
     * @throws \UserReportException
     */
    private function templateGroupCollectFilledElements($max, $htmlDelete, array &$json) {

        $record = $this->store->getStore(STORE_RECORD); // current values
        if ($record === false || count($record) === 0) {
            return '';
        }
        $default = $this->store->getStore(STORE_TABLE_DEFAULT); // current defaults

        // evaluate FE_VALUE on all templateGroup FormElements: After this call, in case of non-primary FEs, FE_VALUE
        // might contain an array of values for all copies of the current FE.
        $maxForeignRecords = $this->templateGroupDoValue();

        $lastFilled = 0; // Marker if there is at least one element per copy who is filled.
        $feSpecNativeCopy = array();
        for ($ii = 1; $ii <= $max; $ii++) {

            // Per copy, iterate over all templateGroup FormElements.
            foreach ($this->feSpecNative as $fe) {

                $fe[FE_NAME] = str_replace(FE_TEMPLATE_GROUP_NAME_PATTERN, $ii, $fe[FE_NAME]);
                $fe[FE_LABEL] = str_replace(FE_TEMPLATE_GROUP_NAME_PATTERN, $ii, $fe[FE_LABEL]);
                $fe[FE_NOTE] = str_replace(FE_TEMPLATE_GROUP_NAME_PATTERN, $ii, $fe[FE_NOTE]);
                $fe[FE_TG_INDEX] = $ii;
                $columnName = $fe[FE_NAME];

                // Column of primary table?
                if (isset($record[$columnName])) {
                    if ($record[$columnName] != $default[$columnName]) {
                        $lastFilled = max($ii, $lastFilled);
                    }

                } else {
                    $lastFilled = max($maxForeignRecords, $lastFilled);
                    if (is_array($fe[FE_VALUE]) && isset($fe[FE_VALUE][$ii - 1])) {
                        $fe[FE_VALUE] = current($fe[FE_VALUE][$ii - 1]); // replace array with current value
                    }
//                    $fe[FE_TEMPLATE_GROUP_CURRENT_IDX] = $ii;
                }

                $feSpecNativeCopy[$ii - 1][] = $fe; // Build array with current copy of templateGroup.
            }

            // Append $htmlDelete on the last element of all copies, but not the first.
            if ($ii > 1) {
                // Count defined FormElements in the current templateGroup
                $last = count($feSpecNativeCopy[$ii - 1]) - 1;
                // Append 'delete' Button at the note of the last element
                $feSpecNativeCopy[$ii - 1][$last][FE_NOTE] .= $htmlDelete;
            }
        }

        // Nothing found: return
        if (count($feSpecNativeCopy) == 0) {
            return '';
        }

        $feSpecNativeSave = $this->feSpecNative;

        $lastFilled = min($lastFilled, $max); // It's possible (external records) that there are more fetched values than the maximum allows - skip those.
        $html = '';
        for ($ii = 0; $ii < $lastFilled; $ii++) {
            $this->feSpecNative = $feSpecNativeCopy[$ii];
            $htmlCopy = $this->elements($this->store->getVar(SIP_RECORD_ID, STORE_SIP), FORM_ELEMENTS_NATIVE, 0, $json);
            $htmlCopy = Support::wrapTag('<div class="qfq-line">', $htmlCopy);
            $html .= $htmlCopy;
        }

        $this->feSpecNative = $feSpecNativeSave;

        return $html;
    }


    /**
     * Evaluate for all FormElements of the current templateGroup the field FE_VALUE.
     * If the specific FormElement is not a real column of the primary table, than the value is probably a
     * '{{!SELECT ...' statement, that one will be fired. In case of an non-primary FE, the result array are the
     * values for the copies of the specific FE.
     *
     * @return int   Count of records in FormElement[FE_VALUE] over all FormElements.
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    private function templateGroupDoValue() {

        // Fire 'value' statement
        $tgMax = 0;
        foreach ($this->feSpecNative as $key => $fe) {

            // Preparation for Log, Debug
            $this->store->setVar(SYSTEM_FORM_ELEMENT, Logger::formatFormElementName($fe), STORE_SYSTEM);
            $this->store->setVar(SYSTEM_FORM_ELEMENT_ID, $fe[FE_ID], STORE_SYSTEM);

            $this->feSpecNative[$key][FE_VALUE] = $this->evaluate->parse($fe[FE_VALUE]);

            if (is_array($this->feSpecNative[$key][FE_VALUE])) {
                $cnt = count($this->feSpecNative[$key][FE_VALUE]);

                $tgMax = max($cnt, $tgMax);
            }
        }

        $this->store->setVar(SYSTEM_FORM_ELEMENT, '', STORE_SYSTEM);
        $this->store->setVar(SYSTEM_FORM_ELEMENT_ID, 0, STORE_SYSTEM);

        return $tgMax;
    }

}